--[[
Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 05.05.2020
Revision: FS22-01

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Mod idea / Concept:
Johnny Vee and GtX

Important:
Not to be added to any mods / maps or modified from its current release form.
No modifications may be made to this script, including conversion to other game versions without written permission from GtX | Andy

Darf nicht zu Mods / Maps hinzugefügt oder von der aktuellen Release-Form geändert werden.
Ohne schriftliche Genehmigung von GtX | Andy dürfen keine Änderungen an diesem Skript vorgenommen werden, einschließlich der Konvertierung in andere Spielversionen
]]

MultiplayerVehicleKeys = {}

MultiplayerVehicleKeys.MOD_NAME = g_currentModName or ""
MultiplayerVehicleKeys.BASE_DIRECTORY = g_currentModDirectory or ""

MultiplayerVehicleKeys.SPEC_NAME = string.format("spec_%s.multiplayerVehicleKeys", g_currentModName)
MultiplayerVehicleKeys.INGAME_NOTIFICATION = {
    0.0227,
    0.5346,
    0.8519,
    1.0
}

local EMPTY_TABLE = {}
local playerUniqueUserId = getUniqueUserId()

function MultiplayerVehicleKeys.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Enterable, specializations) and
        SpecializationUtil.hasSpecialization(Motorized, specializations) and
        not SpecializationUtil.hasSpecialization(Locomotive, specializations)
end

function MultiplayerVehicleKeys.initSpecialization()
    local schemaSavegame = Vehicle.xmlSchemaSavegame
    local savegameKey = string.format("vehicles.vehicle(?).%s.multiplayerVehicleKeys", MultiplayerVehicleKeys.MOD_NAME)

    schemaSavegame:register(XMLValueType.STRING, savegameKey .. "#uniqueUserId", "Unique user ID")
    schemaSavegame:register(XMLValueType.STRING, savegameKey .. "#lastNickname", "Last nickname the user had set", "")
end

function MultiplayerVehicleKeys.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "setIsTabbable", MultiplayerVehicleKeys.setIsTabbable)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanMotorRun", MultiplayerVehicleKeys.getCanMotorRun)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getMotorNotAllowedWarning", MultiplayerVehicleKeys.getMotorNotAllowedWarning)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanStartAIVehicle", MultiplayerVehicleKeys.getCanStartAIVehicle)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "showInfo", MultiplayerVehicleKeys.showInfo)
end

function MultiplayerVehicleKeys.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad",                 MultiplayerVehicleKeys)
    SpecializationUtil.registerEventListener(vehicleType, "onLoadFinished",         MultiplayerVehicleKeys)
    SpecializationUtil.registerEventListener(vehicleType, "onPreDelete",            MultiplayerVehicleKeys)
    SpecializationUtil.registerEventListener(vehicleType, "onReadStream",           MultiplayerVehicleKeys)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteStream",          MultiplayerVehicleKeys)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick",           MultiplayerVehicleKeys)
    SpecializationUtil.registerEventListener(vehicleType, "onLeaveVehicle",         MultiplayerVehicleKeys)
    SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", MultiplayerVehicleKeys)
end

function MultiplayerVehicleKeys.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "addKeyOwner", MultiplayerVehicleKeys.addKeyOwner)
    SpecializationUtil.registerFunction(vehicleType, "removeKeyOwner", MultiplayerVehicleKeys.removeKeyOwner)
    SpecializationUtil.registerFunction(vehicleType, "setKeyOwner", MultiplayerVehicleKeys.setKeyOwner)
    SpecializationUtil.registerFunction(vehicleType, "getHasKeyOwner", MultiplayerVehicleKeys.getHasKeyOwner)
    SpecializationUtil.registerFunction(vehicleType, "getIsKeyOwner", MultiplayerVehicleKeys.getIsKeyOwner)
    SpecializationUtil.registerFunction(vehicleType, "getIsVehicleKeyAvailable", MultiplayerVehicleKeys.getIsVehicleKeyAvailable)
end

function MultiplayerVehicleKeys:onLoad(savegame)
    self.spec_enterableMultiplayerKeys = self[MultiplayerVehicleKeys.SPEC_NAME]

    local spec = self.spec_enterableMultiplayerKeys
    local customEnvironment = self.customEnvironment

    spec.keyOwnerUniqueUserId = ""
    spec.keyOwnerLastNickname = ""

    spec.isTabbable = self:getIsTabbable()

    spec.keyOwnerInfo = g_i18n:getText("info_keyOwner", customEnvironment)
    spec.holdingKeysInfo = g_i18n:getText("info_holdingKeys", customEnvironment)
    spec.holdingKeysTitle = g_i18n:getText("info_holdingKeysTitle", customEnvironment)
    spec.keysReturnedInfo = g_i18n:getText("info_keysReturned", customEnvironment)

    spec.actionRemoveKey = g_i18n:getText("action_removeVehicleKey", customEnvironment)
    spec.actionReturnKey = g_i18n:getText("action_returnVehicleKey", customEnvironment)

    spec.invalidUserWarning = string.format("%s\n%s", g_i18n:getText("warning_youDontHaveAccessToThis"), spec.keyOwnerInfo)
    spec.holdingKeysWarning = string.format("%s\n%s:", g_i18n:getText("warning_actionNotAllowedNow"), spec.holdingKeysTitle)
end

function MultiplayerVehicleKeys:onLoadFinished(savegame)
    if savegame ~= nil then
        local key = string.format("%s.%s.multiplayerVehicleKeys", savegame.key, MultiplayerVehicleKeys.MOD_NAME)
        local uniqueUserId = savegame.xmlFile:getValue(key .. "#uniqueUserId")

        if uniqueUserId ~= nil then
            local lastNickname = savegame.xmlFile:getValue(key .. "#lastNickname", "")

            if playerUniqueUserId == uniqueUserId then
                self:setKeyOwner(uniqueUserId, g_currentMission.playerNickname or lastNickname, false)
            else
                self:setKeyOwner(uniqueUserId, lastNickname, false)
            end
        end
    end
end

function MultiplayerVehicleKeys:onPreDelete()
    self:removeKeyOwner(false, true)

    if g_multiplayerVehicleKeys ~= nil then
        g_multiplayerVehicleKeys.currentVehicleKeys[self] = nil
    end
end

function MultiplayerVehicleKeys:onReadStream(streamId, connection)
    self.spec_enterableMultiplayerKeys.isTabbable = streamReadBool(streamId)

    if streamReadBool(streamId) then
        local uniqueUserId = streamReadString(streamId)
        local lastNickname = streamReadString(streamId)

        self:setKeyOwner(uniqueUserId, lastNickname, false)
    end
end

function MultiplayerVehicleKeys:onWriteStream(streamId, connection)
    local spec = self.spec_enterableMultiplayerKeys

    streamWriteBool(streamId, spec.isTabbable)

    if streamWriteBool(streamId, spec.keyOwnerUniqueUserId ~= "") then
        streamWriteString(streamId, spec.keyOwnerUniqueUserId)
        streamWriteString(streamId, spec.keyOwnerLastNickname or "Unknown Player")
    end
end

function MultiplayerVehicleKeys:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    if self.isClient and isActiveForInputIgnoreSelection then
        local spec = self.spec_enterableMultiplayerKeys
        local actionEvent = spec.actionEvents[InputAction.VEHICLE_KEY_INTERACT]

        if actionEvent ~= nil then
            if self:getIsMotorStarted() then
                g_inputBinding:setActionEventTextPriority(actionEvent.actionEventId, GS_PRIO_LOW)
            else
                g_inputBinding:setActionEventTextPriority(actionEvent.actionEventId, GS_PRIO_VERY_HIGH)
            end

            g_inputBinding:setActionEventText(actionEvent.actionEventId, spec.keyOwnerUniqueUserId == "" and spec.actionRemoveKey or spec.actionReturnKey)
        end
    end
end

function MultiplayerVehicleKeys:onLeaveVehicle()
    local blinkingWarning = g_currentMission.hud.blinkingWarning

    if blinkingWarning ~= nil and blinkingWarning.duration > 0 then
        blinkingWarning.duration = -1
    end
end

function MultiplayerVehicleKeys:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
    local spec = self.spec_enterableMultiplayerKeys

    if isActiveForInputIgnoreSelection then
        self:clearActionEventsTable(spec.actionEvents)

        if (spec.keyOwnerUniqueUserId == "" or spec.keyOwnerUniqueUserId == playerUniqueUserId) and self:getIsActiveForInput(true, true) then
            local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.VEHICLE_KEY_INTERACT, self, MultiplayerVehicleKeys.actionEventVehicleKeyInteract, false, true, false, true, nil)

            g_inputBinding:setActionEventText(actionEventId, spec.keyOwnerUniqueUserId == "" and spec.actionRemoveKey or spec.actionReturnKey)
            g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_LOW)
        end
    end
end

function MultiplayerVehicleKeys:saveToXMLFile(xmlFile, key, usedModNames)
    local spec = self.spec_enterableMultiplayerKeys

    if spec.keyOwnerUniqueUserId ~= "" then
        xmlFile:setValue(key .. "#uniqueUserId", spec.keyOwnerUniqueUserId)
        xmlFile:setValue(key .. "#lastNickname", spec.keyOwnerLastNickname or "Unknown Player")
    end
end

function MultiplayerVehicleKeys:addKeyOwner(userId, showNotification, noEventSend)
    local user = g_currentMission.userManager:getUserByUserId(userId)

    if user ~= nil then
        MultiplayerVehicleKeysSetOwnerEvent.sendEvent(self, userId, showNotification, noEventSend)

        self:setKeyOwner(user:getUniqueUserId(), user:getNickname(), showNotification)
    end
end

function MultiplayerVehicleKeys:removeKeyOwner(showNotification, noEventSend)
    local spec = self.spec_enterableMultiplayerKeys

    MultiplayerVehicleKeysSetOwnerEvent.sendEvent(self, nil, showNotification, noEventSend)

    if spec.keyOwnerUniqueUserId == playerUniqueUserId then
        g_multiplayerVehicleKeys.currentVehicleKeys[self] = nil

        if showNotification then
            g_currentMission.hud:addSideNotification(MultiplayerVehicleKeys.INGAME_NOTIFICATION, string.format(spec.keysReturnedInfo, tostring(self:getFullName())), 4000)
        end
    end

    spec.keyOwnerUniqueUserId = ""
    spec.keyOwnerLastNickname = ""
    spec.isKeyOwner = false

    if self.isClient and self:getIsEntered() then
        self:requestActionEventUpdate()
    end

    self:setIsTabbable(spec.isTabbable, true) -- Restore original setting
end

function MultiplayerVehicleKeys:setKeyOwner(uniqueUserId, nickname, showNotification)
    local spec = self.spec_enterableMultiplayerKeys

    spec.keyOwnerUniqueUserId = uniqueUserId or ""
    spec.keyOwnerLastNickname = nickname or "Unknown Player"
    spec.isKeyOwner = uniqueUserId == playerUniqueUserId

    if spec.isKeyOwner then
        g_multiplayerVehicleKeys.currentVehicleKeys[self] = self

        if showNotification then
            g_currentMission.hud:addSideNotification(MultiplayerVehicleKeys.INGAME_NOTIFICATION, string.format(spec.holdingKeysInfo, tostring(self:getFullName())), 4000)
        end
    end

    self:setIsTabbable(spec.isTabbable and spec.isKeyOwner, true) -- Only for key owner
end

function MultiplayerVehicleKeys:actionEventVehicleKeyInteract(actionName, inputValue, callbackState, isAnalog)
    if self:getIsEntered() and g_multiplayerVehicleKeys ~= nil then
        local currentVehicleKeys = g_multiplayerVehicleKeys:getCurrentKeys() or EMPTY_TABLE

        if currentVehicleKeys[self] then
            self:removeKeyOwner(true)
        else
            local numKeys = table.size(currentVehicleKeys)

            if numKeys < g_multiplayerVehicleKeys:getKeyLimit() then
                self:addKeyOwner(g_currentMission.playerUserId, true)
            else
                local warning = string.format("%s %d/%d \n", self.spec_enterableMultiplayerKeys.holdingKeysWarning, numKeys, g_multiplayerVehicleKeys:getKeyLimit())

                for _, vehicle in pairs (currentVehicleKeys) do
                    warning = string.format("%s\n%s", warning, vehicle:getFullName())
                end

                g_currentMission:showBlinkingWarning(warning, 4000)
            end
        end
    end
end

function MultiplayerVehicleKeys:getHasKeyOwner()
    return self.spec_enterableMultiplayerKeys.keyOwnerUniqueUserId ~= ""
end

function MultiplayerVehicleKeys:getIsKeyOwner(uniqueUserId)
    return self.spec_enterableMultiplayerKeys.keyOwnerUniqueUserId == (uniqueUserId or playerUniqueUserId)
end

function MultiplayerVehicleKeys:getIsVehicleKeyAvailable()
    local spec = self.spec_enterableMultiplayerKeys

    if spec.keyOwnerUniqueUserId == "" or spec.keyOwnerUniqueUserId == playerUniqueUserId then
        return true
    end

    if self:getIsAIActive() then
        return true
    end

    local user = g_currentMission.userManager:getUserByUserId(self.spec_enterable.controllerUserId)

    if user ~= nil then
        if spec.keyOwnerUniqueUserId == user:getUniqueUserId() then
            return true
        end

        if user:getIsMasterUser() then
            return g_currentMission.userManager:getUserByUniqueId(spec.keyOwnerUniqueUserId) == nil
        end
    end

    return false
end

function MultiplayerVehicleKeys:showInfo(superFunc, box)
    local spec = self.spec_enterableMultiplayerKeys

    if spec.keyOwnerUniqueUserId ~= "" and spec.keyOwnerLastNickname ~= "" then
        box:addLine(spec.holdingKeysTitle, spec.keyOwnerLastNickname or "Unknown Player")
    end

    superFunc(self, box)
end

function MultiplayerVehicleKeys:getCanMotorRun(superFunc)
    return self:getIsVehicleKeyAvailable() and superFunc(self)
end

function MultiplayerVehicleKeys:getMotorNotAllowedWarning(superFunc)
    if not self:getIsVehicleKeyAvailable() then
        local spec = self.spec_enterableMultiplayerKeys

        return string.format(spec.invalidUserWarning, spec.keyOwnerLastNickname or "Unknown Player")
    end

    return superFunc(self)
end

function MultiplayerVehicleKeys:getCanStartAIVehicle(superFunc)
    local spec = self.spec_enterableMultiplayerKeys

    if spec.keyOwnerUniqueUserId ~= "" and spec.keyOwnerUniqueUserId ~= playerUniqueUserId then
        return false
    end

    return superFunc(self)
end

function MultiplayerVehicleKeys:setIsTabbable(superFunc, isTabbable, ignoreMultiplayerVehicleKey)
    superFunc(self, isTabbable)

    -- Extra parameter not designed for use outside of this spec...
    -- It allows other mods to set tabbable state correctly using this base game function
    -- Also allows me to override and restore state when keys are used while update any other spec overwriting the function
    if not ignoreMultiplayerVehicleKey then
        self.spec_enterableMultiplayerKeys.isTabbable = isTabbable
    end
end
